from __future__ import annotations

from pathlib import Path

from jinja2.environment import Environment
from jinja2.loaders import FileSystemLoader


class Scaffolder:
    def __init__(self, source_dir: Path, target_dir: Path, render_context: dict | None = None):
        self.source_dir = source_dir
        self.target_dir = target_dir
        self.render_context = render_context or {}
        self.template_suffix = ".jinja2"

        self.env = Environment(
            autoescape=False,
            keep_trailing_newline=True,
            trim_blocks=True,
            lstrip_blocks=True,
            block_end_string="%]",
            block_start_string="[%",
            comment_end_string="#]",
            comment_start_string="[#",
            variable_end_string="]]",
            variable_start_string="[[",
            loader=FileSystemLoader(self.source_dir),
        )

    def run(self):
        self._render_folder(self.source_dir)

    def _render_folder(self, src_abspath: Path) -> None:
        """Recursively render a folder.

        Args:
            src_path:
                Folder to be rendered. It must be an absolute path within
                the template.
        """
        assert src_abspath.is_absolute()
        src_relpath = src_abspath.relative_to(self.source_dir)
        dst_relpath = self._render_path(src_relpath)
        if dst_relpath is None:
            return
        if not self._render_allowed(dst_relpath, is_dir=True):
            return
        dst_abspath = Path(self.target_dir, dst_relpath)
        dst_abspath.mkdir(parents=True, exist_ok=True)
        for file in src_abspath.iterdir():
            if file.is_dir():
                self._render_folder(file)
            else:
                self._render_file(file)

    def _render_path(self, relpath: Path) -> Path | None:
        """Render one relative path.

        Args:
            relpath:
                The relative path to be rendered. Obviously, it can be templated.
        """
        is_template = relpath.name.endswith(self.template_suffix)
        # If there is a sibling file in the same dir that is templated, skip the current file
        templated_sibling = self.source_dir / f"{relpath}{self.template_suffix}"
        # With an empty suffix, the templated sibling always exists.
        if templated_sibling.exists():
            return None
        if self.template_suffix and is_template:
            relpath = relpath.with_suffix("")
        rendered_parts = []
        for part in relpath.parts:
            # Skip folder if any part is rendered as an empty string
            part = self._render_string(part)
            if not part:
                return None
            # restore part to be just the end leaf
            rendered_parts.append(part)
        result = Path(*rendered_parts)
        if not is_template:
            templated_sibling = self.source_dir / f"{result}{self.template_suffix}"
            if templated_sibling.exists():
                return None
        return result

    def _render_file(self, src_abspath: Path) -> None:
        """Render one file.

        Args:
            src_abspath:
                The absolute path to the file that will be rendered.
        """
        assert src_abspath.is_absolute()
        src_relpath = src_abspath.relative_to(self.source_dir).as_posix()
        src_renderpath = src_abspath.relative_to(self.source_dir)
        dst_relpath = self._render_path(src_renderpath)
        if dst_relpath is None:
            return
        if src_abspath.name.endswith(self.template_suffix):
            try:
                tpl = self.env.get_template(src_relpath)
            except UnicodeDecodeError:
                if self.template_suffix:
                    # suffix is not empty, re-raise
                    raise
                # suffix is empty, fallback to copy
                new_content = src_abspath.read_bytes()
            else:
                new_content = tpl.render(**self.render_context).encode()
        else:
            new_content = src_abspath.read_bytes()
        dst_abspath = Path(self.target_dir, dst_relpath)
        src_mode = src_abspath.stat().st_mode
        if not self._render_allowed(dst_relpath, expected_contents=new_content):
            return
        dst_abspath.parent.mkdir(parents=True, exist_ok=True)
        dst_abspath.write_bytes(new_content)
        dst_abspath.chmod(src_mode)

    def _render_string(self, string: str) -> str:
        """Render one templated string.

        Args:
            string:
                The template source string.
        """
        tpl = self.env.from_string(string)
        return tpl.render(**self.render_context)

    def _render_allowed(
        self,
        dst_relpath: Path,
        is_dir: bool = False,
        expected_contents: bytes | Path = b"",
    ) -> bool:
        """Determine if a file or directory can be rendered.

        Args:
            dst_relpath:
                Relative path to destination.
            is_dir:
                Indicate if the path must be treated as a directory or not.
            expected_contents:
                Used to compare existing file contents with them. Allows to know if
                rendering is needed.
        """
        assert not dst_relpath.is_absolute()
        assert not expected_contents or not is_dir, "Dirs cannot have expected content"
        dst_abspath = Path(self.target_dir, dst_relpath)
        if not is_dir and dst_abspath.exists():
            previous_content: bytes = dst_abspath.read_bytes()
            return previous_content != expected_contents
        if is_dir:
            return True
        return True



from __future__ import annotations

from pathlib import Path

from pharaoh.plugins.api import L1Template, impl

DEFAULT_ASSET_TEMPLATE_MAPPING = {
    ".html": "iframe",  # todo: use raw_html if this is not a full HTML file, iframe otherwise
    ".rst": "raw_rst",
    ".txt": "raw_txt",
    ".svg": "image",
    ".png": "image",
    ".jpg": "image",
    ".jpeg": "image",
    ".gif": "image",
    ".md": "markdown",
}


@impl
def pharaoh_collect_l1_templates():
    return [
        L1Template(
            name="pharaoh.default_project",
            path=Path(__file__).parent / "level_1" / "pharaoh_sphinx_project",
        ),
        L1Template(
            name="pharaoh.empty",
            path=Path(__file__).parent / "level_1" / "empty",
        ),
        L1Template(
            name="pharaoh.report_info",
            path=Path(__file__).parent / "level_1" / "report_info",
        ),
    ]


@impl
def pharaoh_collect_resource_types():
    from pharaoh.assetlib.resource import CustomResource, FileResource

    return [FileResource, CustomResource]


@impl(tryfirst=True)
def pharaoh_collect_default_settings():
    return Path(__file__).with_name("default_settings.yaml")


@impl(tryfirst=True)
def pharaoh_find_asset_render_template(template_name: str) -> Path:
    template_dir = Path(__file__).with_name("asset_templates")
    assert template_dir.exists()
    template = template_dir / f"{template_name}.rst.jinja2"
    if not template.exists():
        msg = f"Cannot find template {template_name!r} in {template_dir}!"
        raise LookupError(msg)
    return template


@impl(tryfirst=True)
def pharaoh_get_asset_render_template_mappings():
    return DEFAULT_ASSET_TEMPLATE_MAPPING



# mypy: disable-error-code="method-assign"
from __future__ import annotations

import uuid
from contextlib import contextmanager
from pathlib import Path

try:
    import pandas as pd

    PANDAS_AVAIL = True
except (ImportError, ModuleNotFoundError):
    PANDAS_AVAIL = False

from pharaoh import project
from pharaoh.assetlib.context import context_stack
from pharaoh.assetlib.util import parse_signature

if PANDAS_AVAIL:
    # memorize the unpatched functions
    vanilla_df_to_html = pd.DataFrame.to_html

    @contextmanager
    def patch():
        """
        A context manager that patches all functions of pandas relevant to export html when executed in a pharaoh
        context. When the program exits the patch-context, the patching will be undone.

        The main goals:
            - retain original function call signature
            - alter storage location of figures to the pharaoh build folder
            - inject table generation parameters via config file
        """

        # switching out the original with the patched functions
        if pd.DataFrame.to_html == vanilla_df_to_html:
            pd.DataFrame.to_html = patched_df_to_html

        yield

        # undo the patching
        if pd.DataFrame.to_html != vanilla_df_to_html:
            pd.DataFrame.to_html = vanilla_df_to_html

    def patched_df_to_html(*args, **kwargs):
        """
        This is the patched version of ``pandas.DataFrame.to_html(...)`` function.

        Changes:
            - if a storage location is passed in (optional) it changes the destination to be inside pharaoh's asset
              build folder
            - takes kwargs from the pharaoh config file and updates the kwargs passed into the function

        docs: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.to_html.html
        """
        params = parse_signature(obj=vanilla_df_to_html, args=args, kwargs=kwargs, fromclass=True)
        buf = params["buf"]

        if buf is None:
            return  # when to_html() is called without a destination (optional), there's nothing to patch

        active_app = project.get_project()

        # relocating the saved file by patching the buf parameter
        file_path = active_app._build_asset_filepath(Path(buf))
        with context_stack.new_context(
            asset={
                "user_filepath": str(buf),
                "file": str(file_path),
                "name": file_path.name,
                "stem": file_path.stem,
                "suffix": file_path.suffix,
                "template": "datatable",
            }
        ):
            params["buf"] = str(file_path)
            params["table_id"] = "T_" + uuid.uuid4().hex[:5]  # make sure id does not start with a number
            pandas_config = active_app.get_setting("toolkits.pandas.to_html")
            params.update(pandas_config)

            vanilla_df_to_html(**params)
            context_stack.dump(file_path)

else:
    # if the package is not installed there is nothing to patch, so we silently do nothing when patch() is called.

    @contextmanager
    def patch():
        yield



from __future__ import annotations

import importlib
import importlib.util
import os
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pluggy
import wrapt

import pharaoh.plugins.core_plugin.plugin
from pharaoh.plugins import NAME, spec

if TYPE_CHECKING:
    import types
    from collections.abc import Iterable

    import click

    from pharaoh.api import PharaohProject, PharaohSphinx, Resource
    from pharaoh.plugins.template import L1Template


_plugin_result_cache: dict[str, dict] = {}


@wrapt.decorator
def cached(wrapped, instance, args, kwargs):
    func = wrapped.__name__
    key = (args, tuple(kwargs))
    try:
        return _plugin_result_cache[func][key]
    except KeyError:
        result = wrapped(*args, **kwargs)
        _plugin_result_cache.setdefault(func, {})
        _plugin_result_cache[func][key] = result
        return result


def clear_cache():
    _plugin_result_cache.clear()


class PharaohPluginManager:
    """
    Responsible for discovering Pharaoh plugins via namespace packages and PHARAOH_PLUGINS environment variable.
    """

    def __init__(self):
        self.pm = pluggy.PluginManager("pharaoh")

        def before(hook_name, hook_impls, kwargs):
            ...
            # log.debug(f"Plugins: Executing hook {hook_name!r} from plugins " +
            #           ", ".join([impl.plugin_name for impl in hook_impls]))

        def after(outcome, hook_name, hook_impls, kwargs): ...

        self.pm.add_hookcall_monitoring(before=before, after=after)
        self.pm.add_hookspecs(spec)
        self.reload_plugins()

    def reload_plugins(self):
        for name, plugin in self.pm.list_name_plugin():
            self.pm.unregister(plugin, name)

        clear_cache()

        self.pm.load_setuptools_entrypoints(NAME)
        for plugin in self._discover_plugin_modules():
            self.pm.register(plugin)
        self.pm.register(pharaoh.plugins.core_plugin.plugin, name="pharaoh_core")
        self.pm.check_pending()

    @contextmanager
    def disable_plugins(self, *names: str):
        """
        Temporarily disable loaded plugins::

            with PM.disable_plugins("plugin_abc", "..."):
                ...

        :param names: Names of the loaded plugins
        """
        plugins = [(name, self.pm.get_plugin(name)) for name in names]
        for _, plugin in plugins:
            if plugin is not None:
                self.pm.unregister(plugin)
        clear_cache()
        try:
            yield
        finally:
            clear_cache()
            for name, plugin in plugins:
                if plugin is not None:
                    self.pm.register(plugin, name)

    def _discover_plugin_modules(self) -> Iterable[types.ModuleType]:
        """
        Discovers all plugins.

        :return: An iterable of plugins
        """
        for path_item in os.environ.get("PHARAOH_PLUGINS", "").strip("\"'").split(";"):
            path_item = path_item.strip().strip("\"'")
            if not path_item:
                continue
            path = Path(path_item) / "plugin.py"
            if not path.exists():
                msg = f"Cannot find plugin path {path} from environment variable PHARAOH_PLUGINS!"
                raise FileNotFoundError(msg)

            spec = importlib.util.spec_from_file_location(path.stem, str(path))
            if not spec:
                msg = f"Could not get module spec from {path}!"
                raise OSError(msg)
            plugin_module = importlib.util.module_from_spec(spec)
            if not plugin_module:
                msg = f"Plugin {path} is not a valid Python module!"
                raise OSError(msg)
            if not spec.loader:  # pragma: no cover
                msg = f"Could not get module loader for {path}"
                raise OSError(msg)
            spec.loader.exec_module(plugin_module)

            yield plugin_module

    @cached
    def pharaoh_collect_l1_templates(self) -> dict[str, L1Template]:
        """
        Returns a mapping of first-level templates
        """
        return {t.name: t for templates in self.pm.hook.pharaoh_collect_l1_templates() for t in templates}

    @cached
    def pharaoh_collect_l2_templates(self) -> list[Path]:
        return [path for paths in self.pm.hook.pharaoh_collect_l2_templates() for path in paths]

    @cached
    def pharaoh_collect_resource_types(self) -> dict[str, type[Resource]]:
        resources = [resource for resources in self.pm.hook.pharaoh_collect_resource_types() for resource in resources]
        return {type_.__name__: type_ for type_ in resources}

    def pharaoh_asset_gen_prepare_resources(self, project: PharaohProject, resources: dict[str, Resource]):
        return self.pm.hook.pharaoh_asset_gen_prepare_resources(project=project, resources=resources)

    def pharaoh_configure_sphinx(self, project: PharaohProject, config: dict[str, Any], confdir: Path):
        return self.pm.hook.pharaoh_configure_sphinx(project=project, config=config, confdir=confdir)

    def pharaoh_collect_default_settings(self) -> list[Path | dict]:
        return self.pm.hook.pharaoh_collect_default_settings()

    def pharaoh_add_cli_commands(self, cli: click.Group):
        return self.pm.hook.pharaoh_add_cli_commands(cli=cli)

    def pharaoh_build_finished(self, project, exc_info):
        return self.pm.hook.pharaoh_build_finished(project=project, exc_info=exc_info)

    def pharaoh_modify_gitignore(self, gitignore_lines: list[str]):
        return self.pm.hook.pharaoh_modify_gitignore(gitignore_lines=gitignore_lines)

    def pharaoh_project_created(self, project: PharaohProject):
        return self.pm.hook.pharaoh_project_created(project=project)

    def pharaoh_asset_gen_started(self, project: PharaohProject):
        return self.pm.hook.pharaoh_asset_gen_started(project=project)

    def pharaoh_build_started(self, project: PharaohProject, builder: str):
        return self.pm.hook.pharaoh_build_started(project=project, builder=builder)

    def pharaoh_sphinx_app_inited(self, app: PharaohSphinx):
        return self.pm.hook.pharaoh_sphinx_app_inited(app=app)

    @cached
    def pharaoh_find_asset_render_template(self, template_name: str) -> list[Path]:
        templates = self.pm.hook.pharaoh_find_asset_render_template(template_name=template_name)
        return list({Path(t) for t in templates})

    @cached
    def pharaoh_get_asset_render_template_mappings(self) -> dict[str, str]:
        mapping = {}
        for m in self.pm.hook.pharaoh_get_asset_render_template_mappings():
            mapping.update(m)
        return mapping


PM = PharaohPluginManager()



from __future__ import annotations

import hashlib
import json
import shutil
from pathlib import Path
from typing import TYPE_CHECKING

import omegaconf

from pharaoh.log import log

from .util import obj_groupby

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator


class AssetFileLinkBrokenError(LookupError):
    pass


class Asset:
    """
    Holds information about a generated asset.

    :ivar id: An MD5 hash of the asset's filename, prefixed with "__ID__".
        Since a unique suffix is included in the filename, this ID hash is also unique.
        Can be used to quickly find this Asset instance.
    :ivar Path infofile: Absolute path to the ``*.assetinfo`` file
    :ivar Path assetfile: Absolute path to the actual asset file
    :ivar omegaconf.DictConfig context: The content of *infofile* parsed into a OmegaConf dict.
    """

    def __init__(self, info_file: Path):
        assert info_file.suffix == ".assetinfo"
        self.id: str = "__ID__" + hashlib.md5(bytes(info_file.name, "utf-8")).hexdigest()
        self.infofile: Path = info_file
        self.context = omegaconf.OmegaConf.create(json.loads(self.infofile.read_text()))
        for file in self.infofile.parent.glob(f"{self.infofile.stem}*"):
            if file.suffix != ".assetinfo":
                self.assetfile: Path = file
                break
        else:
            msg = f"There is no asset for inventory file {self.infofile}!"
            raise AssetFileLinkBrokenError(msg)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f"Asset[{self.infofile.stem}]"

    def __eq__(self, other):
        if isinstance(other, Asset):
            return self.id == other.id and self.infofile == other.infofile and self.assetfile == other.assetfile
        raise NotImplementedError

    def __hash__(self):
        return hash(self.infofile.name)

    def __lt__(self, other):
        if isinstance(other, Asset):
            return self.infofile < other.infofile
        raise NotImplementedError

    def copy_to(self, target_dir: Path) -> Path:
        """
        Copy the asset plus info-file.

        :param target_dir: The target directory to copy to. Will be created if it does not exist.
        :returns: True if files were copied, False otherwise (files already exist)
        """
        target_dir.mkdir(exist_ok=True, parents=True)
        target_info_file = target_dir / self.infofile.name
        if target_info_file.exists():
            return target_dir / self.assetfile.name

        log.debug(f"Copying asset {self} to build directory")
        shutil.copy(self.infofile, target_info_file)

        if Path(self.assetfile).is_file():
            return Path(shutil.copy(self.assetfile, target_dir))

        if Path(self.assetfile).is_dir():
            shutil.copytree(self.assetfile, target_dir / self.assetfile.name)
            return target_dir / self.assetfile.name

        raise NotImplementedError

    def read_json(self) -> dict:
        """
        Reads the file using a JSON parser.
        """
        if self.assetfile.suffix.lower() != ".json":
            msg = "Can only read .json files!"
            raise Exception(msg)
        return json.loads(self.assetfile.read_text("utf-8"))

    def read_yaml(self) -> dict:
        """
        Reads the file using a YAML parser.
        """
        import yaml

        if self.assetfile.suffix.lower() not in (".yaml", ".yml"):
            msg = "Can only read .yaml/.yml files!"
            raise Exception(msg)
        with open(self.assetfile, encoding="utf-8") as fp:
            return yaml.safe_load(fp)

    def read_text(self, encoding: str = "utf-8") -> str:
        """
        Reads the file as text
        """
        return self.assetfile.read_text(encoding)

    def read_bytes(self) -> bytes:
        """
        Reads the file as bytes
        """
        return self.assetfile.read_bytes()


class AssetFinder:
    def __init__(self, lookup_path: Path):
        """
        A class for discovering and searching generated assets.

        An instance of this class will be created by the Pharaoh project, where ``lookup_path`` will be set to
        ``report_project/.asset_build``.

        :param lookup_path: The root directory to look for assets. It will be searched recursively for assets.
        """
        self._lookup_path = lookup_path
        self._assets: dict[str, list[Asset]] = {}
        self.discover_assets()

    def discover_assets(self, components: list[str] | None = None) -> dict[str, list[Asset]]:
        """
        Discovers all assets by recursively searching for ``*.assetinfo`` files and stores
        the collection as instance variable (`_assets`).

        :param components: A list of components to search for assets.
            If None (the default), all components will be searched.
        :return: A dictionary that maps component names to a list of :class:`Asset` instances.
        """
        if isinstance(components, list) and len(components):
            for component in components:
                self._assets[component] = [Asset(file) for file in (self._lookup_path / component).glob("*.assetinfo")]
        else:
            self._assets.clear()
            for asset in (Asset(file) for file in self._lookup_path.glob("*/*.assetinfo")):
                component = asset.assetfile.parent.name
                if component not in self._assets:
                    self._assets[component] = []
                self._assets[component].append(asset)

        return self._assets

    def search_assets(self, condition: str, components: str | Iterable[str] | None = None) -> list[Asset]:
        """
        Searches already discovered assets (see :func:`discover_assets`) that match a condition.

        :param condition: A Python expression that is evaluated using the content of the ``*.assetinfo`` JSON file
            as namespace. If the evaluation returns a truthy result, the asset is returned.

            Refer to :ref:`this example assetinfo file <example_asset_info>` to see the available default namespace.

            Example::

                # All HTML file where the "label" metadata ends with "_plot"
                finder.search_assets('asset.suffix == ".html" and label.endswith("_plot")')

        :param components: A list of component names to search. If None (the default), all components will be searched.
        :return: A list of assets whose metadata match the condition.
        """
        if not condition.strip():
            return []

        code = compile(condition, "<string>", "eval")
        found = []

        for asset in self.iter_assets(components):
            try:
                result = eval(code, {}, asset.context)
            except Exception:
                result = False
            if result:
                found.append(asset)

        return AssetFinder.sort_assets(found)

    def iter_assets(self, components: str | Iterable[str] | None = None) -> Iterator[Asset]:
        """
        Iterates over all discovered assets.

        :param components: A list of component names to search. If None (the default), all components will be searched.
        :return: An iterator over all discovered assets.
        """
        if not self._assets:
            self.discover_assets()

        if isinstance(components, str):
            components = [components]
        components = components or list(self._assets.keys())
        for component in components:
            if component in self._assets:
                yield from self._assets[component]

    def get_asset_by_id(self, id: str) -> Asset | None:
        """
        Returns the corresponding :class:`Asset` instance for a certain ID.

        :param id: The ID of the asset to return
        :return: An :class:`Asset` instance if found, None otherwise.
        """
        for asset in self.iter_assets():
            if asset.id == id:
                return asset
        return None

    @staticmethod
    def sort_assets(assets: list[Asset]) -> list[Asset]:
        """
        Sorts a list of assets by their generation order.

        :param assets: A list of :class:`Asset` instances to sort
        """

        def sort_key(asset):
            try:
                return asset.context.asset.index
            except AttributeError:
                return 0

        return sorted(assets, key=sort_key)


def asset_groupby(
    seq: Iterable[Asset], key: str, sort_reverse: bool = False, default: str | None = None
) -> dict[str, list[Asset]]:
    """
    Groups an iterable of Assets by a certain metadata key.

    During build-time rendering this function will be available as Jinja global function
    ``asset_groupby`` and alias ``agroupby``.

    Example:

    .. code-block:: none

        We have following 4 assets (simplified notation of specified metadata):
        Asset[a="1", b="3"]
        Asset[a="1", c="4"]
        Asset[a="2", b="3"]
        Asset[a="2", c="4"]

        Grouping by "a":
            asset_groupby(assets, "a")
        will yield
            {
                "1": [Asset[a="1", b="3"], Asset[a="1", c="4"]],
                "2": [Asset[a="2", b="3"], Asset[a="2", c="4"]],
            }


        Grouping by "b" and default "default":
            asset_groupby(assets, "b", default="default")
        will yield
            {
                "3":       [Asset[a="1", b="3"], Asset[a="2", b="3"]],
                "default": [Asset[a="1", c="4"], Asset[a="2", c="4"]],
            }


    :param seq: The iterable of assets to group
    :param key: The nested attribute to use for grouping, e.g. "A.B.C"
    :param sort_reverse: Reverse-sort the keys in the returned dictionary
    :param default: Sort each item, where "key" is not an existing attribute, into this default group
    :return: A dictionary that maps the group names (values of A.B.C) to a list of items out of the input iterable
    """
    return obj_groupby(seq=seq, key=key, sort_reverse=sort_reverse, attr="context", default=default)



from __future__ import annotations

import concurrent.futures
import contextlib
import io
import json
import logging.handlers
import multiprocessing
import os
import re
import shutil
import traceback
from functools import partial
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Union

import pharaoh.log
from pharaoh import project
from pharaoh.assetlib import patches
from pharaoh.assetlib.context import context_stack
from pharaoh.assetlib.finder import Asset
from pharaoh.templating.second_level.sphinx_ext.asset_tmpl import find_asset_template
from pharaoh.util.contextlib_chdir import chdir
from pharaoh.util.json_encoder import CustomJSONEncoder

if TYPE_CHECKING:
    from collections.abc import Iterable
    from queue import Queue

log = pharaoh.log.log

PathLike = Union[str, Path]


def generate_assets(project_root: Path, asset_src: Path, component_name: str = "", mp_log_queue: Queue | None = None):
    # Since this function is always called in a process by generate_assets_parallel,
    # we need to remove at least all file handlers so child-processes don't log to the same file
    # as the parent process, otherwise race conditions may occur.
    # So we just remove all handlers and add a QueueHandler
    # to send all log records to the parent in order to handle them.
    if mp_log_queue is not None:  # pragma: no cover
        for hdl in log.handlers:
            log.removeHandler(hdl)
        log.addHandler(logging.handlers.QueueHandler(mp_log_queue))

    # Also forbid the project instance to add loggers we just removed
    proj = project.PharaohProject(project_root=project_root, logging_add_filehandler=False)
    context_stack.reset()

    try:
        script_path = asset_src.relative_to(project_root).as_posix()
    except ValueError:
        script_path = asset_src.as_posix()

    if asset_src.suffix.lower() == ".py":
        script_ignore_pattern = proj.get_setting("asset_gen.script_ignore_pattern")

        with (
            patches.patch_3rd_party_libraries(),
            context_stack.new_context(
                context_name="generate_assets",
                asset={
                    "script_name": asset_src.name,
                    "script_path": asset_src,
                    "index": 0,
                    "component_name": component_name,
                },
            ),
        ):
            code = asset_src.read_text(encoding="utf-8")

            first_line = code.split("\n", maxsplit=1)[0].strip()
            if re.fullmatch(script_ignore_pattern, asset_src.name) or re.fullmatch(
                r"^# *pharaoh?: *ignore *", first_line, re.IGNORECASE
            ):
                log.info(f"Ignoring file {script_path}")
                return

            log.info(f"Generating assets from script {script_path!r}...")
            asset_module = module_from_file(asset_src)

            WAVEWATSON_LEGACY_INPLACE = os.environ.get("WAVEWATSON_LEGACY_INPLACE")
            os.environ["WAVEWATSON_LEGACY_INPLACE"] = "0"
            try:
                run_module(asset_module, code)
            except Exception as e:
                msg = (
                    f"An exception was raised when executing module "
                    f"{str(asset_src)!r}:\n\n{e}\n\nTraceback:\n{traceback.format_exc()}"
                )
                raise Exception(msg) from None
            finally:
                if WAVEWATSON_LEGACY_INPLACE is None:
                    del os.environ["WAVEWATSON_LEGACY_INPLACE"]
                else:
                    os.environ["WAVEWATSON_LEGACY_INPLACE"] = WAVEWATSON_LEGACY_INPLACE

    elif asset_src.suffix.lower() == ".ipynb":
        # import locally, otherwise Sphinx autodoc on pharaoh.assetlib.api fails because of nbformat package and this
        # issue: https://github.com/sphinx-doc/sphinx/issues/11662
        import nbformat
        from nbconvert.preprocessors import CellExecutionError, ExecutePreprocessor

        log.info(f"Generating assets from notebook {script_path!r}...")

        with open(asset_src) as file:
            nb = nbformat.read(file, as_version=4)
        initial_node = nbformat.notebooknode.from_dict(
            {
                "cell_type": "code",
                "execution_count": None,
                "id": "000000",
                "outputs": [],
                "metadata": {},
                "source": f"""
import os
from pharaoh.api import PharaohProject
from pharaoh.assetlib.api import metadata_context
from pharaoh.assetlib.patches import patch_3rd_party_libraries

proj = PharaohProject(project_root="{project_root.as_posix()}")
patcher = patch_3rd_party_libraries()
patcher.__enter__()
metadata_context(
    context_name="generate_assets",
    asset=dict(
        script_name="{asset_src.name}",
        script_path="{asset_src.as_posix()}",
        component_name="{component_name}",
        index=0
    )
).activate()

os.environ["WAVEWATSON_LEGACY_INPLACE"] = "0"
os.environ["JPY_SESSION_NAME"] = "{asset_src.as_posix()}"
"""[1:-1],
            }
        )
        # Remark: JPY_SESSION_NAME is normally set by Jupyter runtime. In our case we use the nbconvert preprocessor
        # so the variable is not set. So we have do set it manually to the current notebook path in order for the
        # pharaoh.assetlib.api.get_current_component function to be able to find the "executing" script..

        nb.cells.insert(0, initial_node)

        ep = ExecutePreprocessor(timeout=600)
        subdir = component_name or "default"
        try:
            ep.preprocess(nb, {"metadata": {"path": str(asset_src.parent)}})

            completed_notebooks_path = proj.asset_build_dir / "completed_notebooks" / subdir / asset_src.name
            completed_notebooks_path.parent.mkdir(parents=True, exist_ok=True)
            with open(completed_notebooks_path, "w", encoding="utf-8") as file:
                nbformat.write(nb, file)
        except CellExecutionError as e:
            failed_notebooks_path = proj.asset_build_dir / "failed_notebooks" / subdir / asset_src.name
            failed_notebooks_path.parent.mkdir(parents=True, exist_ok=True)
            with open(failed_notebooks_path, "w", encoding="utf-8") as file:
                nbformat.write(nb, file)
            msg = (
                f"An exception was raised when executing notebook '{asset_src.stem}': {e}\n"
                f"Check the notebook for errors/traces: {failed_notebooks_path}"
            )

            # Remove all 7-bit C1 ANSI sequences for better readability
            ansi_escape = re.compile(
                r"""
                \x1B  # ESC
                (?:   # 7-bit C1 Fe (except CSI)
                    [@-Z\\-_]
                |     # or [ for CSI, followed by a control sequence
                    \[
                    [0-?]*  # Parameter bytes
                    [ -/]*  # Intermediate bytes
                    [@-~]   # Final byte
                )
            """,
                re.VERBOSE,
            )
            msg = ansi_escape.sub("", msg)

            raise Exception(msg) from None


def generate_assets_parallel(
    project_root: PathLike, asset_sources: Iterable[tuple[str, Path]], workers: str | int = "auto"
):
    project_root = Path(project_root)
    if isinstance(workers, int):
        workers = max(1, workers)
    if isinstance(workers, str):
        if workers.lower() == "auto":
            workers = multiprocessing.cpu_count()
        else:
            msg = "Argument worker may only be an integer number or the string 'auto'!"
            raise ValueError(msg)

    log.info(f"Executing asset generation with {workers} worker processes")
    generate_asset_partial = partial(generate_assets, project_root=project_root)

    mp_manager = multiprocessing.Manager()
    mp_log_queue = mp_manager.Queue(-1)

    results = []
    # The queue listener collects all log records handled via the queue handler (defined inside generate_assets)
    # in order to log them in the parent process
    ql = logging.handlers.QueueListener(mp_log_queue, *log.handlers, respect_handler_level=True)
    ql.start()
    with concurrent.futures.ProcessPoolExecutor(max_workers=workers) as executor:
        futures_map = {
            executor.submit(
                generate_asset_partial, asset_src=asset_source, component_name=component_name, mp_log_queue=mp_log_queue
            ): asset_source
            for component_name, asset_source in asset_sources
        }
        for future in concurrent.futures.as_completed(futures_map.keys()):
            result = None
            try:
                result = future.result()
                results.append((futures_map[future], result))
            except SystemExit as e:
                if e.code == 0:
                    results.append((futures_map[future], result))
                else:
                    results.append((futures_map[future], e))
            except Exception as e:
                results.append((futures_map[future], e))
    ql.stop()
    return results


def module_from_file(path: str | Path) -> ModuleType:
    """
    Creates a module from file at runtime.

    :param path: Path to the module source code.
    :return: A fresh module
    """

    module_path = Path(path)
    module_name = f"asset_module_{module_path.stem.replace(' ', '_').replace('-', '_')}"
    module = ModuleType(module_name)
    module.__dict__["__file__"] = str(module_path.absolute())
    module.__path__ = [str(module_path.parent)]
    module.__dict__["__name__"] = "__main__"
    module.__dict__["__module_name__"] = module_name

    return module


def run_module(module: ModuleType, code: str):
    """
    Execute the configured source code in a module.

    :param module: A module object.
    :param code: The code to be run inside the module
    """
    compiled_code = compile(code, module.__dict__["__file__"], "exec")
    with chdir(Path(module.__dict__["__file__"]).parent):
        exec(compiled_code, module.__dict__)


def register_asset(
    file: PathLike,
    metadata: dict | None = None,
    template: str | None = None,
    data: io.BytesIO | None = None,
    copy2build: bool = False,
    **kwargs,
) -> Asset | None:
    """
    Register an asset manually. The file will be copied (if data is None and 'file' is a real file) or
    written (if data is given) to the asset build folder of the current Pharaoh project.

    :param file: The filename. Must exist even if data is set, to have a filename to store the asset and to
        automatically determine the template (if not set via template argument).
    :param metadata: Additional metadata to store on the asset.
    :param template: The template used to render the asset. If omitted, it is inferred by the file extension.

        .. seealso:: :ref:`reference/directive:Asset Templates`

    :param data: An io.BytesIO instance. Used if the asset is generated in memory and should be stored to disk by
                 this function.
    :param copy2build: If True, the asset will be copied to the asset build directory,
        even if not referenced in the template.

        Background: Pharaoh stores all assets in the project directory and copies them to the build directory only if
        copy2build is set to True or on-demand by Pharaoh. For example if an HTML file is rendered using an iframe,
        the HTML file has to be copied to the build folder where the iframe can later include it.

    :returns: The file path where the asset will be actually stored
    """
    try:
        active_app = project.get_project()
    except Exception:
        # If there is no Pharaoh application yet, the calling file is presumably executed standalone,
        # so we have to skip exporting any files
        return None

    component_name = kwargs.pop("component", None)
    if kwargs:
        raise Exception("Unknown keyword arguments " + ",".join(kwargs.keys()))

    file = Path(file)
    if not template:
        suffix = file.suffix.lower()
        from pharaoh.plugins.plugin_manager import PM

        template = PM.pharaoh_get_asset_render_template_mappings().get(suffix)

    if template is not None:
        find_asset_template(template)  # will fail if template does not exist

    if template in ("iframe",):
        copy2build = True

    # If this function is used in an asset script that is executed directly, the component name is not added to the
    # metadata context, so we have to find the component via the callstack and pass it to _build_asset_filepath
    from pharaoh.assetlib.api import get_current_component

    if component_name is None:
        try:
            component_name = get_current_component()
        except LookupError:  # raised if method is not executed from inside a component
            msg = (
                "When register_asset is called outside a component of a Pharaoh project, keyword argument "
                "'component' must be set!"
            )
            raise Exception(msg) from None
    asset_file_path = active_app._build_asset_filepath(file, component_name)

    metadata = metadata or {}
    metadata.pop("context_name", None)
    metadata.pop("asset", None)
    with context_stack.new_context(
        context_name="manual_registry",
        asset={
            "user_filepath": str(file),
            "file": str(asset_file_path),
            "name": asset_file_path.name,
            "stem": asset_file_path.stem,
            "suffix": asset_file_path.suffix,
            "template": template,
            "copy2build": copy2build,
        },
        **metadata,
    ):
        with contextlib.suppress(LookupError):
            # If asset scripts are executed directly, this context does not exist
            context_stack.get_parent_context(name="generate_assets")["asset"]["index"] += 1
        if isinstance(data, io.BytesIO):
            with open(asset_file_path, "wb") as fp:
                fp.write(bytes(data.getbuffer()))
        else:
            if file.is_file():
                shutil.copy(file.absolute(), asset_file_path)
            elif file.is_dir():
                shutil.copytree(file, asset_file_path)
            else:
                msg = f"{file} does not exist!"
                raise FileNotFoundError(msg)
        info_file = context_stack.dump(asset_file_path)
    return Asset(info_file)


def register_templating_context(name: str, context: str | Path | dict | list, metadata: dict | None = None, **kwargs):
    """
    Register a data context for the build-time templating stage.
    The data may be given directly as dict/list or via a json or yaml file.

    This function is designed to be used within asset scripts, to easily register data you extract from resources
    for the templating process.

    Example::

        from pharaoh.assetlib.api import register_templating_context

        register_templating_context(name="foo", context={"bar": "baz"})
        # will be accessed like this: {{ ctx.local.foo.bar.baz }}

    :param name: The name under which the data context is available inside Jinja templates.
        Access like this (name: mycontext)::

            {% set mycontext = ctx.local.mycontext %}

    :param context: Either a str or :external:class:`Path <pathlib.Path>` instance pointing to a json or yaml
        file, or a dict or list. All data must contain only json-compatible types, otherwise the data cannot be stored.
    :param metadata: The given context will be internally registered as an asset with following metadata:
        ``dict(pharaoh_templating_context=name, **metadata)``
    :param kwargs: Keyword arguments that are mostly (except ``component``) passed to ``json.dumps(...)``, in case
        ``context`` is a dict or list.
    """
    component = kwargs.pop("component", None)
    metadata = metadata or {}
    metadata.pop("pharaoh_templating_context", None)
    kwargs.setdefault("cls", CustomJSONEncoder)

    if not name:
        msg = "name must be a non-empty string!"
        raise ValueError(msg)

    if isinstance(context, (str, Path)):
        file = Path(context)
        if file.suffix.lower() not in (".json", ".yaml"):
            msg = "If context is a file path, it's suffix must be either .json or .yaml!"
            raise ValueError(msg)
        register_asset(file, metadata=dict(pharaoh_templating_context=name, **metadata), component=component)
    elif isinstance(context, (list, dict)):
        data = io.BytesIO(json.dumps(context, **kwargs).encode("utf-8"))
        register_asset(
            "pharaoh_templating_context.json",
            metadata=dict(pharaoh_templating_context=name, **metadata),
            data=data,
            component=component,
        )
    else:
        msg = f"Unsupported type {type(context)}!"
        raise TypeError(msg)



from __future__ import annotations

import uuid
from functools import partial
from pathlib import Path

from pharaoh.assetlib.finder import asset_groupby


def raise_helper(msg):
    """
    Raises an Exception with a certain message. The actual name of the Jinja global function is ``raise``.
    """
    raise Exception(msg)


def heading(text: str, level: int) -> str:
    """
    Renders a heading for a certain level.

    Shortcuts for this function are automatically generated with names h1...h7.

    Examples::

        {{ heading("Page Title", 1) }}
        {{ h2("Sub-Title") }}

    :param text: The heading
    :param level: The heading level. From 1 (top-level) to 7.
    """
    default = {1: "#", 2: "*", 3: "=", 4: "-", 5: "^", 6: '"', 7: "~"}
    if isinstance(level, (float, int)):
        if int(level) not in default:
            msg = "Default heading characters only available from index 1 to 7!"
            raise LookupError(msg)
        character = default[level]
    else:
        character = str(level)
        if character not in default.values():
            raise ValueError("character must be an integer from 1-7 or one of " + ",".join(default.values()))

    text = text.strip().replace("\r", "").replace("\n", "")
    if len(text) < 2:
        msg = "Heading text must be at least 2 characters"
        raise ValueError(msg)
    return f"{text}\n{character * len(text)}"


def rand_id(chars: int | None = None) -> str:
    """
    Returns a unique id for usage in HTML as it is made sure it starts with no number
    """
    id = "i" + uuid.uuid4().hex
    return id[:chars]


def read_text(file) -> str:
    """
    Reads a file content using utf-8 encoding.
    """
    file = Path(file)
    return file.read_text("utf-8")


def hrule():
    """
    Renders a 1px gray horizontal rule using a raw html directive.
    """
    return '.. raw:: html\n\n    <hr style="height:1px;border-width:0;color:gray;background-color:gray">\n\n'


def fglob(pattern: str, root: str = ".") -> list[Path]:
    """
    Returns a list of relative paths of all files relative to a root directory that match a certain pattern.

    Examples::

        {% set index_files1 = fglob("index_*.rst") %}
        {% set index_files2 = fglob("subdir/index_*.rst") %}

    :param pattern: A glob pattern to match files.
    :param root: The root directory for discovering files and anchor for relative paths.
        Default is the current working directory (the parent directory of the currently rendered file).
    """
    rootpath = Path(root).absolute()
    matches = rootpath.glob(pattern)
    return [file.relative_to(rootpath) for file in matches]


def assert_true(statement: bool, message: str = ""):
    """
    Wrapper for Python's assert builtin.

    :param statement: A boolean statement
    :param message: The message for the AssertionError, if statement is False.
    """
    if message:
        assert statement, message
    else:
        assert statement


env_globals = {
    "assert_true": assert_true,
    "fglob": fglob,
    "hrule": hrule,
    "raise": raise_helper,
    "heading": heading,
    "rand_id": rand_id,
    "read_text": read_text,
    "agroupby": asset_groupby,
    "asset_groupby": asset_groupby,
}
# Shortcuts for headings
for i in range(1, 8):
    env_globals[f"h{i}"] = partial(heading, level=i)  # type: ignore[assignment]


# Only document the functions defined in here
to_document = {
    k: func for k, func in env_globals.items() if hasattr(func, "__module__") and func.__module__ == __name__
}



"""
Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

from __future__ import annotations

from pathlib import Path
from typing import Any, Callable

from pharaoh.api import PharaohProject

THISDIR = Path(__file__).parent
proj = PharaohProject(project_root=THISDIR.parent)
extensions: list[str] = []

# Dynamic updates of local Sphinx configuration with defaults from Pharaoh project.
locals().update(proj.get_default_sphinx_configuration(THISDIR))

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# Your own extensions if you like
extensions.extend([])

# A list of template lookup directories relative to this file's parent directory
pharaoh_jinja_templates = ["user_templates"]

# The default user context that may be used in all templates by accessing via {{ ctx.user.XYZ }}
# E.g. pharaoh_jinja_context = dict(a=1, b=dict(c=2))  -->  {{ ctx.user.a }} or {{ ctx.user.b.c }}
pharaoh_jinja_context: dict[str, Any] = {}

# Add your custom Jinja filters here. Maps filter names to filter functions.
# See https://jinja.palletsprojects.com/en/3.1.x/templates/#filters
pharaoh_jinja_filters: dict[str, Callable] = {}

# Add your custom Jinja globals here. Maps global names to global functions.
pharaoh_jinja_globals: dict[str, Callable] = {}

# Add your custom Jinja tests here. Maps test names to test functions.
# See https://jinja.palletsprojects.com/en/3.1.x/templates/#tests
pharaoh_jinja_tests: dict[str, Callable] = {}



from __future__ import annotations

import abc
import copy
import glob
from pathlib import Path
from typing import ClassVar

import attrs
import natsort
import omegaconf
from attrs.validators import deep_iterable, in_, instance_of, min_len

__all__ = ["CustomResource", "FileResource", "LocalResource", "Resource", "TransformedResource"]


@attrs.define
class Resource:
    """
    The base class for all resources. Provides serialization/deserialization methods.
    """

    _registry: ClassVar[dict] = {}

    alias: str = attrs.field(validator=(instance_of(str), min_len(1)))
    _cachedir: str = attrs.field(validator=instance_of(str), kw_only=True, default="")

    @staticmethod
    def from_dict(definition: dict | omegaconf.DictConfig) -> Resource:
        from pharaoh.plugins.plugin_manager import PM

        if isinstance(definition, omegaconf.DictConfig):
            definition: dict = omegaconf.OmegaConf.to_container(definition, resolve=True)
        else:
            definition: dict = copy.deepcopy(definition)
        resource_class = definition.pop("__class__")
        registry = PM.pharaoh_collect_resource_types()
        if resource_class not in registry:
            msg = (
                f"No such resource type {resource_class!r} from definition of resource {definition}. "
                f"Available types are {','.join(Resource._registry.keys())!r}!"
            )
            raise LookupError(msg)
        return registry[resource_class](**definition)

    def to_dict(self):
        dikt = dict(**attrs.asdict(self, filter=lambda attr, value: not attr.name.startswith("_")))
        dikt["__class__"] = self.__class__.__name__
        return dikt


@attrs.define
class LocalResource(abc.ABC, Resource):
    """
    Represents a resource that is located on the user's hard-drive.
    """

    @abc.abstractmethod
    def locate(self) -> Path:
        """
        Returns the path to the resource file.
        """
        raise NotImplementedError


@attrs.define
class TransformedResource(LocalResource, abc.ABC):
    """
    A resource that is depending on another resource and maybe transforms it dynamically into another resource.
    """

    sources: list[str] = attrs.field(
        validator=(
            deep_iterable(
                member_validator=instance_of(str),
                iterable_validator=instance_of(list),
            ),
            min_len(1),
        )
    )

    @abc.abstractmethod
    def transform(self, resources: dict[str, Resource]):
        """
        Executed by Pharaoh before asset generation.
        Transforms the linked resource into a new one.
        The resulting resource may be cached or recreated each time.
        """
        raise NotImplementedError


@attrs.define
class CustomResource(Resource):
    """
    A resource that can hold arbitrary information. It may be used as a temporary replacement for
    more specific resources that are not yet implemented.
    For example, it could hold database connection properties or other custom data.

    Example::

        CustomResource(alias="foo", traits=dict(a=1, b=[2, 3], c=dict(d=[4, 5])))


    :ivar traits: A dict with arbitrary data.
    """

    traits: dict


@attrs.define
class FileResource(LocalResource):
    """
    A resource that matches files/directories using a wildcard pattern.

    Example::

        FileResource(alias="mycsvfile", pattern="C:/temp/*.csv", sort="ascending")

    :ivar pattern: A pathname pattern.
        The pattern may contain simple shell-style wildcards.
        Filenames starting with a dot are special cases that are not matched by '*' and '?' patterns.
    :ivar sort: descending (default) or ascending.
        Sorting is done using the natsort library to sort filesystem paths naturally. Example::

            0.txt, 01.txt, 1.txt, 10.txt, 2.txt  # default sorting (ascending)
            0.txt, 01.txt, 1.txt, 2.txt, 10.txt  # natural sorting (ascending)

    """

    pattern: str | Path = attrs.field(converter=str)

    sort: str = attrs.field(validator=in_(("ascending", "descending")), default="descending")

    def locate(self) -> Path:
        """
        Returns the first match for the file pattern, depending on the chosen sort order.
        """
        return self.first_match()

    def first_match(self) -> Path:
        """
        Returns the first match for the file pattern, depending on the chosen sort order.
        """
        return self.get_files()[0]

    def last_match(self) -> Path:
        """
        Returns the last match for the file pattern, depending on the chosen sort order.
        """
        return self.get_files()[-1]

    def get_match(self, index) -> Path:
        """
        Returns the n-th match for the file pattern, depending on the chosen sort order.
        """
        return self.get_files()[index]

    def get_files(self, recursive: bool = True) -> list[Path]:
        """
        Returns all matches for the file pattern, depending on the chosen sort order.

        :param recursive:   If recursive is true, the pattern '**' will match any files and
                            zero or more directories and subdirectories.
        """
        files = [Path(p).resolve() for p in glob.glob(self.pattern, recursive=recursive)]  # type: ignore[type-var]
        if not files:
            msg = f"No files found for pattern {self.pattern!r}"
            raise FileNotFoundError(msg)
        sort = {"ascending": False, "descending": True}
        return natsort.natsorted(files, reverse=sort[self.sort])



from __future__ import annotations

import typing as t
from pathlib import Path

from sphinx.application import Sphinx

if t.TYPE_CHECKING:  # We get circular imports otherwise
    from pharaoh.project import PharaohProject
    from pharaoh.templating.second_level.template_env import PharaohTemplateEnv


class PharaohSphinx(Sphinx):
    pharaoh_proj: PharaohProject
    """
    Set by entrypoint of Sphinx plugin ...asset_ext.setup()
    """

    pharaoh_te: PharaohTemplateEnv
    """
    Set by entrypoint of Sphinx plugin ...jinja_ext.setup()
    """

    seen_warnings: t.ClassVar = set()
    """
    Set of warnings that have been seen already. Used to avoid duplicate warnings.
    Added by PharaohAssetDirective.
    """

    @property
    def assets_dir(self):
        return Path(self.builder.outdir) / "pharaoh_assets"



from __future__ import annotations

from typing import TYPE_CHECKING

import pharaoh
from pharaoh.templating.second_level.template_env import PharaohTemplateEnv

if TYPE_CHECKING:
    from pharaoh.sphinx_app import PharaohSphinx


def setup(app: PharaohSphinx):
    app.add_config_value("pharaoh_jinja_templates", None, "env")
    app.add_config_value("pharaoh_jinja_context", None, "env")
    app.add_config_value("pharaoh_jinja_filters", None, "env")
    app.add_config_value("pharaoh_jinja_tests", None, "env")
    app.add_config_value("pharaoh_jinja_globals", None, "env")

    app.pharaoh_te = PharaohTemplateEnv()
    app.connect("config-inited", app.pharaoh_te.sphinx_config_inited_hook)
    app.connect("builder-inited", app.pharaoh_te.sphinx_builder_inited_hook)
    app.connect("source-read", app.pharaoh_te.sphinx_source_read_hook)
    app.connect("include-read", app.pharaoh_te.sphinx_include_read_hook)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
        "version": pharaoh.__version__,
    }
